'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.feeLevels = undefined;
exports.estimateFee = estimateFee;
exports.getFees = getFees;

var _jquery = require('jquery');

var _jquery2 = _interopRequireDefault(_jquery);

var _currency = require('./currency');

var _serverConnect = require('./serverConnect');

var _app = require('../app');

var _app2 = _interopRequireDefault(_app);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

var feeLevels = exports.feeLevels = ['PRIORITY', 'NORMAL', 'ECONOMIC'];

var cacheExpires = 1000 * 60 * 5;
var estimateFeeCache = new Map();
var watchingTransactions = false;

function onSocket(e) {
  if (e.jsonData.wallet && !e.jsonData.wallet.height) {
    estimateFeeCache.clear();
  }
}

/**
 * Since a new transactions will potentially affect inputs and change the estimated fee,
 * we'll clear the cache when any new transactions come.
 */
function watchTransactions() {
  if (watchingTransactions) return;
  watchingTransactions = true;

  var socket = (0, _serverConnect.getSocket)();
  if (!socket) return;

  socket.on('message', onSocket);

  // in case we connect to a new server
  _serverConnect.events.on('connected', function (e) {
    socket.off(null, onSocket);
    e.socket.on('message', onSocket);
  });
}

/**
 * Will estimate a fee based on the provided amount, fee level and the current state
 * of the wallet.
 *
 * @param {string} feeLevel - The fee level
 * @param {amount} number - The amount of the transaction in the servers currency -
 *   not in base units. (e.g. for BTC, provide a BTC amount, not Satoshi)
 * @return {object} An jQuery promise which on success will resolve with the fee
 *   in Bitcoin. If the call fails, the deferred will fail and pass on the args the
 *   xhr fail handler receives.
 */
function estimateFee(feeLevel, amount) {
  if (feeLevels.indexOf(feeLevel) === -1) {
    throw new Error('feelevel must be one of ' + feeLevels.join(', '));
  }

  if (typeof amount !== 'number') {
    throw new Error('Please provide an amount as a number.');
  }

  watchTransactions();

  var deferred = void 0;
  var cacheKey = feeLevel + '-' + amount;
  var cached = estimateFeeCache.get(cacheKey);

  if (cached) {
    if (cached && Date.now() - cached.createdAt < cacheExpires) {
      deferred = cached.deferred;
    } else {
      // cache is expired
      estimateFeeCache.delete(cacheKey);
    }
  }

  if (!deferred) {
    deferred = _jquery2.default.Deferred();

    estimateFeeCache.set(cacheKey, {
      deferred: deferred,
      createdAt: Date.now()
    });

    var queryArgs = 'feeLevel=' + feeLevel + '&amount=' + (0, _currency.decimalToInteger)(amount, _app2.default.serverConfig.cryptoCurrency);
    _jquery2.default.get(_app2.default.getServerUrl('wallet/estimatefee/?' + queryArgs)).done(function () {
      var _deferred;

      for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
        args[_key] = arguments[_key];
      }

      (_deferred = deferred).resolve.apply(_deferred, [(0, _currency.integerToDecimal)(args[0].estimatedFee, _app2.default.serverConfig.cryptoCurrency)].concat(_toConsumableArray(args.slice(1))));
    }).fail(function (xhr) {
      var _deferred2;

      for (var _len2 = arguments.length, args = Array(_len2 > 1 ? _len2 - 1 : 0), _key2 = 1; _key2 < _len2; _key2++) {
        args[_key2 - 1] = arguments[_key2];
      }

      (_deferred2 = deferred).reject.apply(_deferred2, [xhr].concat(args));

      var knownReasons = ['ERROR_INSUFFICIENT_FUNDS', 'ERROR_DUST_AMOUNT'];

      // don't cache calls that failed with an unknown reason
      if (xhr.responseJSON && knownReasons.indexOf(xhr.responseJSON.reason) === -1) {
        estimateFeeCache.delete(cacheKey);
      }
    });
  }

  return deferred.promise();
}

var getFeesCache = {};

/**
 * Will call the fees api ('wallet/fees') on the server.
 *
 * @return {object} An jQuery promise which on success will resolve with the fees
 *   per byte in Satoshi for each fee level. If the call fails, the deferred will
 *   fail and pass on the args the xhr fail handler receives.
 */
function getFees() {
  var deferred = void 0;

  if (getFeesCache.deferred && Date.now() - getFeesCache.createdAt < cacheExpires) {
    deferred = getFeesCache.deferred;
  }

  if (!deferred) {
    deferred = _jquery2.default.Deferred();

    getFeesCache = {
      deferred: deferred,
      createdAt: Date.now()
    };

    _jquery2.default.get(_app2.default.getServerUrl('wallet/fees')).done(function () {
      var _deferred3;

      return (_deferred3 = deferred).resolve.apply(_deferred3, arguments);
    }).fail(function () {
      var _deferred4;

      (_deferred4 = deferred).reject.apply(_deferred4, arguments);
      getFeesCache = {};
    });
  }

  return deferred.promise();
}