'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.btcSymbol = exports.events = undefined;

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

exports.getEvents = getEvents;
exports.NoExchangeRateDataError = NoExchangeRateDataError;
exports.UnrecognizedCurrencyError = UnrecognizedCurrencyError;
exports.decimalToInteger = decimalToInteger;
exports.integerToDecimal = integerToDecimal;
exports.formatPrice = formatPrice;
exports.formatCurrency = formatCurrency;
exports.fetchExchangeRates = fetchExchangeRates;
exports.getExchangeRate = getExchangeRate;
exports.getExchangeRates = getExchangeRates;
exports.convertCurrency = convertCurrency;
exports.convertAndFormatCurrency = convertAndFormatCurrency;
exports.getCurrencyValidity = getCurrencyValidity;
exports.renderFormattedCurrency = renderFormattedCurrency;
exports.renderPairedCurrency = renderPairedCurrency;

var _underscore = require('underscore');

var _underscore2 = _interopRequireDefault(_underscore);

var _app = require('../app');

var _app2 = _interopRequireDefault(_app);

var _jquery = require('jquery');

var _jquery2 = _interopRequireDefault(_jquery);

var _bitcoinConvert = require('bitcoin-convert');

var _bitcoinConvert2 = _interopRequireDefault(_bitcoinConvert);

var _number = require('./number');

var _backbone = require('backbone');

var _currencies = require('../data/currencies');

var _cryptoCurrencies = require('../data/cryptoCurrencies');

var _cryptoListingCurrencies = require('../data/cryptoListingCurrencies');

var _loadTemplate = require('../utils/loadTemplate');

var _loadTemplate2 = _interopRequireDefault(_loadTemplate);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

var events = _extends({}, _backbone.Events);

exports.events = events;

// friendlier for circular dependancies

function getEvents() {
  return events;
}

var btcSymbol = exports.btcSymbol = '₿';

function NoExchangeRateDataError(message) {
  this.message = message || 'Missing exchange rate data';
  this.name = 'NoExchangeRateDataError';
  this.stack = new Error().stack;
}

NoExchangeRateDataError.prototype = Object.create(Error.prototype);
NoExchangeRateDataError.prototype.constructor = NoExchangeRateDataError;

function UnrecognizedCurrencyError(message) {
  this.message = message || 'The currency is not recognized.';
  this.name = 'UnrecognizedCurrencyError';
  this.stack = new Error().stack;
}

UnrecognizedCurrencyError.prototype = Object.create(Error.prototype);
UnrecognizedCurrencyError.prototype.constructor = UnrecognizedCurrencyError;

/**
 * Converts the amount from a decimal to an integer. If the currency code is a crypto currency,
 * it will convert to its base units.
 */
function decimalToInteger(amount, currency) {
  var options = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

  if (typeof amount !== 'number') {
    throw new Error('Please provide an amount as a number.');
  }

  if (typeof currency !== 'string') {
    throw new Error('Please provide a currency as a string.');
  }

  var opts = _extends({
    returnUndefinedOnError: true
  }, options);

  var curData = (0, _currencies.getCurrencyByCode)(currency);
  var returnVal = void 0;

  if (!curData) {
    if (!opts.returnUndefinedOnError) {
      throw new UnrecognizedCurrencyError(currency + ' is not a recognized currency.');
    }
  } else {
    if ((0, _cryptoCurrencies.getCurrencyByCode)(currency)) {
      returnVal = Math.round(amount * curData.baseUnit);
    } else {
      returnVal = Math.round(amount * 100);
    }
  }

  return returnVal;
}

/**
 * Converts the amount from an integer to a decimal, rounding to 2 decimal places. If the
 * currency code is for a crypto currency, it will convert from its base units.
 */
function integerToDecimal(amount, currency) {
  var options = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

  var opts = _extends({
    returnUndefinedOnError: true
  }, options);

  var curData = (0, _currencies.getCurrencyByCode)(currency);
  var returnVal = void 0;

  if (!curData) {
    if (!opts.returnUndefinedOnError) {
      throw new UnrecognizedCurrencyError(currency + ' is not a recognized currency.');
    }
  } else {
    if ((0, _cryptoCurrencies.getCurrencyByCode)(currency)) {
      returnVal = Number((amount / curData.baseUnit).toFixed(curData.maxDisplayDecimals));
    } else {
      returnVal = Number((amount / 100).toFixed(2));
    }
  }

  return returnVal;
}

/**
 * Will take a number and return a string version of the number with the appropriate number of
 * decimal places based on whether the number represents a crypto or fiat price.
 *
 * This differs from formatCurrency in that this does not localize the number at all. It simply
 * returns the value with the appropriate number of decimal place, e.g:
 *
 * formatPrice(123.456, 'USD') // "123.46"
 * formatPrice(123.456, 'BTC')  // "123.45600000"
 *
 * It is more useful for <input>'s because we are not localizing the numbers in them.
 *
 */
function formatPrice(price, currency) {
  if (typeof price !== 'number') {
    throw new Error('Please provide a price as a number');
  }

  if (isNaN(price)) {
    throw new Error('Please provide a price that is not NaN');
  }

  if (typeof currency !== 'string') {
    throw new Error('Please provide a currency as a string');
  }

  var convertedPrice = void 0;
  var cryptoCur = (0, _cryptoCurrencies.getCurrencyByCode)(currency);

  if (cryptoCur) {
    // Format crypto price so it has up to the max decimal places (as specified in the crypto
    // config), but without any trailing zeros
    convertedPrice = (0, _number.upToFixed)(price, cryptoCur.maxDisplayDecimals);
  } else {
    convertedPrice = price.toFixed(2);
  }

  return convertedPrice;
}

/**
 * Will format an amount in the given currency into the format appropriate for the given locale.
 * In many cases, instead of using this method directly, you may want to use
 * renderFormattedCurrency() from this module or its corresponding template helper,
 * formattedCurrency, since those will more robustly handle (via tooltips and icons)
 * unrecognized currency codes and/or conversion problems due to unavailable exchange
 * rate data.
 */
function formatCurrency(amount, currency, options) {
  var opts = _extends({
    locale: _app2.default && _app2.default.localSettings && _app2.default.localSettings.standardizedTranslatedLang() || 'en-US',
    btcUnit: _app2.default && _app2.default.localSettings && _app2.default.localSettings.get('bitcoinUnit') || 'BTC'
  }, options);

  if (typeof amount !== 'number' || isNaN(amount)) {
    return '';
  }

  if (typeof opts.locale !== 'string') {
    throw new Error('Please provide a locale as a string');
  }

  if (typeof currency !== 'string') {
    throw new Error('Please provide a currency as a string');
  }

  var cur = currency.toUpperCase();
  var curData = (0, _currencies.getCurrencyByCode)(cur);

  var formattedCurrency = void 0;
  var cryptoCur = (0, _cryptoCurrencies.getCurrencyByCode)(cur);

  // If we don't recognize the currency, we'll assume it's a crypto
  // listing cur.
  var isCryptoListingCur = (0, _cryptoListingCurrencies.getCurrencies)().includes(cur) || !cryptoCur && !curData;

  if (cryptoCur) {
    var curSymbol = curData.symbol || cur;
    var bitcoinConvertUnit = void 0;
    var amt = amount;

    if (cur === 'BTC' || cur === 'TBTC') {
      switch (opts.btcUnit) {
        case 'MBTC':
          bitcoinConvertUnit = curSymbol = 'mBTC';
          break;
        case 'UBTC':
          bitcoinConvertUnit = curSymbol = 'μBTC';
          break;
        case 'SATOSHI':
          curSymbol = 'sat';
          bitcoinConvertUnit = 'Satoshi';
          break;
        default:
          bitcoinConvertUnit = 'BTC';
      }

      amt = (0, _bitcoinConvert2.default)(amount, 'BTC', bitcoinConvertUnit);
    }

    var formattedAmount = new Intl.NumberFormat(opts.locale, {
      minimumFractionDigits: opts.minDisplayDecimals || curData.minDisplayDecimals,
      maximumFractionDigits: opts.maxDisplayDecimals || curData.maxDisplayDecimals
    }).format(amt);

    var translationSubKey = curSymbol === curData.symbol ? 'curSymbolAmount' : 'curCodeAmount';
    formattedCurrency = _app2.default.polyglot.t('cryptoCurrencyFormat.' + translationSubKey, _defineProperty({
      amount: formattedAmount
    }, curSymbol === curData.symbol ? 'symbol' : 'code', curSymbol));
  } else if (isCryptoListingCur) {
    var _formattedAmount = new Intl.NumberFormat(opts.locale, {
      minimumFractionDigits: opts.minDisplayDecimals || 0,
      maximumFractionDigits: opts.maxDisplayDecimals || 8
    }).format(amount);

    formattedCurrency = _app2.default.polyglot.t('cryptoCurrencyFormat.curCodeAmount', {
      amount: _formattedAmount,
      code: cur.length > 8 ? cur.slice(0, 8) + '\u2026' : cur
    });
  } else {
    formattedCurrency = new Intl.NumberFormat(opts.locale, {
      style: 'currency',
      currency: currency,
      minimumFractionDigits: opts.minDisplayDecimals || 2,
      maximumFractionDigits: opts.maxDisplayDecimals || 2
    }).format(amount);
  }

  return formattedCurrency;
}

var exchangeRates = {};

/**
 * Will fetch exchange rate data from the server. This is already called on an interval via
 * exchangeRateSyncer.js, so it's unlikely you would need to call this method. Instead access
 * cached values via getExchangeRate() or more commonly convertCurrency().
 */
function fetchExchangeRates() {
  var options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

  var xhr = _jquery2.default.get(_app2.default.getServerUrl('ob/exchangerates/'), options).done(function (data) {
    var changed = new Set();

    Object.keys(exchangeRates).forEach(function (cur) {
      if (exchangeRates[cur] !== data[cur]) {
        changed.add(cur);
      }
    });

    Object.keys(data).forEach(function (cur) {
      if (data[cur] !== exchangeRates[cur]) {
        changed.add(cur);
      }
    });

    var changedArray = Array.from(changed);
    var prevExchangeRates = JSON.parse(JSON.stringify(exchangeRates));
    exchangeRates = data;

    if (changed.size) {
      events.trigger('exchange-rate-change', { changed: changedArray });
      changedArray.forEach(function (cur) {
        events.trigger('exchange-rate-change-' + cur, { previous: prevExchangeRates[cur] });
      });
    }
  });

  events.trigger('fetching-exchange-rates', { xhr: xhr });

  return xhr;
}

/**
 * Will return the exchange rate between the server's crypto currency and the given
 * currency.
 */
function getExchangeRate(currency) {
  if (!currency) {
    throw new Error('Please provide a currency.');
  }

  var returnVal = exchangeRates[currency];
  var serverCurrency = (0, _cryptoCurrencies.getServerCurrency)();

  if (serverCurrency.code === currency || serverCurrency.testnetCode === currency) {
    returnVal = 1;
  }

  return returnVal;
}

/**
 * Will return an object containing all the available exchange rates for your servers
 * crypto currency.
 */
function getExchangeRates() {
  return exchangeRates;
}

/**
 * Converts an amount from one currency to another based on exchange rate data.
 */
function convertCurrency(amount, fromCur, toCur) {
  var fromCurCaps = fromCur.toUpperCase();
  var toCurCaps = toCur.toUpperCase();
  var serverCur = (0, _cryptoCurrencies.getServerCurrency)();
  var isFromServerCur = fromCurCaps === serverCur.code || fromCurCaps === serverCur.testnetCode;
  var isToServerCur = toCurCaps === serverCur.code || toCurCaps === serverCur.testnetCode;

  if (typeof amount !== 'number') {
    throw new Error('Please provide an amount as a number');
  }

  if (isNaN(amount)) {
    throw new Error('Please provide an amount that is not NaN');
  }

  if (typeof fromCurCaps !== 'string') {
    throw new Error('Please provide a fromCur as a string');
  }

  if (typeof toCurCaps !== 'string') {
    throw new Error('Please provide a toCur as a string');
  }

  if (fromCurCaps === toCurCaps) {
    return amount;
  }

  if (!isFromServerCur && !exchangeRates[fromCurCaps]) {
    throw new NoExchangeRateDataError('We do not have exchange rate data for ' + fromCurCaps + '.');
  }

  if (!isToServerCur && !exchangeRates[toCurCaps]) {
    throw new NoExchangeRateDataError('We do not have exchange rate data for ' + toCurCaps + '.');
  }

  var fromRate = isFromServerCur ? 1 : getExchangeRate(fromCurCaps);
  var toRate = isToServerCur ? 1 : getExchangeRate(toCurCaps);

  return amount / fromRate * toRate;
}

/**
 * Convenience function to both convert and format a currency amount using convertCurrency()
 * and formatCurrency(). In many cases, instead of using this method directly, you may want
 * to use renderFormattedCurrency() from this module or its corresponding template helper,
 * formattedCurrency, since those will more robustly handle (via tooltips and icons)
 * unrecognized currency codes and/or conversion problems due to unavailable exchange rate data.
 */
function convertAndFormatCurrency(amount, fromCur, toCur) {
  var options = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : {};

  var opts = _extends({
    locale: _app2.default && _app2.default.localSettings && _app2.default.localSettings.standardizedTranslatedLang() || 'en-US',
    btcUnit: _app2.default && _app2.default.localSettings && _app2.default.localSettings.get('bitcoinUnit') || 'BTC',
    skipConvertOnError: true
  }, options);

  var convertedAmt = void 0;
  var outputFormat = toCur;

  try {
    convertedAmt = convertCurrency(amount, fromCur, toCur);
  } catch (e) {
    if (opts.skipConvertOnError) {
      // We'll use an unconverted amount
      convertedAmt = amount;
      outputFormat = fromCur;
    } else {
      throw e;
    }
  }

  return formatCurrency(convertedAmt, outputFormat, _underscore2.default.omit(opts, 'skipConvertOnError'));
}

/**
 * Returns `VALID` if the given currency is valid, otherwise it will return a code
 * indicating why it's not valid.
 */
function getCurrencyValidity(cur) {
  if (typeof cur !== 'string') {
    throw new Error('A currency must be provided as a string.');
  }

  var curData = (0, _currencies.getCurrencyByCode)(cur);
  var returnVal = void 0;

  if (curData) {
    // Determine if the given currency is the one the node is running in
    var isCurServerCurrency = curData.isCrypto && _app2.default.serverConfig.cryptoCurrency === cur || _app2.default.serverConfig.cryptoCurrency === curData.testnetCode;

    if (getExchangeRate(cur) || isCurServerCurrency) {
      returnVal = 'VALID';
    } else {
      returnVal = 'EXCHANGE_RATE_MISSING';
    }
  } else {
    returnVal = 'UNRECOGNIZED_CURRENCY';
  }

  return returnVal;
}

/**
 * Will render a formattedCurrency template. The main function of the template is that it will
 * render a localized price when possible. When it is not possible (e.g. an unrecognized currency),
 * it will render an alert icon with a tooltip containing an explanation (assuming you don't pass in
 * the showTooltipOnUnrecognizedCur option as false).
 */
function renderFormattedCurrency(amount, fromCur, toCur) {
  var options = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : {};

  if (typeof fromCur !== 'string' || !fromCur) {
    throw new Error('Please provide a "from currency" as a string.');
  }

  if (toCur && typeof toCur !== 'string') {
    throw new Error('If providing a "to currency", it must be provided as a string.');
  }

  var result = '';

  (0, _loadTemplate2.default)('components/formattedCurrency.html', function (t) {
    result = t(_extends({
      price: amount,
      fromCur: fromCur,
      toCur: toCur || fromCur
    }, options));
  });

  return result;
}

/**
 * Will render a pairing of currencies, most commonly used to show a crypto currency
 * along with it's fiat equivalent (e.g. $2.33 (0.0002534 BTC)). If it cannot show the
 * "to" currency (e.g. exchange rate data not available), it will just show the "from".
 * If the "from" currency is invalid, it will render an empty string.
 */
function renderPairedCurrency(price, fromCur, toCur) {
  var fromCurValidity = getCurrencyValidity(fromCur);

  if (typeof price !== 'number' || fromCurValidity === 'UNRECOGNIZED_CURRENCY') {
    // Sometimes when prices are in an unsupported currency, they will be
    // saved as empty strings or undefined. We'll ignore those an just render an
    // empty string.
    return '';
  }

  var toCurValidity = getCurrencyValidity(toCur);
  var formattedBase = formatCurrency(price, fromCur);
  var formattedConverted = fromCur === toCur || toCurValidity !== 'VALID' || fromCurValidity !== 'VALID' ? '' : convertAndFormatCurrency(price, fromCur, toCur);

  var result = formattedBase;

  if (formattedConverted !== '') {
    result = _app2.default.polyglot.t('currencyPairing', {
      baseCurValue: formattedBase,
      convertedCurValue: formattedConverted
    });
  }

  return result;
}