'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.events = undefined;

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

exports.getDebugLog = getDebugLog;
exports.getCurrentConnection = getCurrentConnection;
exports.getServer = getServer;
exports.getSocket = getSocket;
exports.disconnect = disconnect;
exports.default = connect;

var _os = require('os');

var _electron = require('electron');

var _underscore = require('underscore');

var _underscore2 = _interopRequireDefault(_underscore);

var _jquery = require('jquery');

var _jquery2 = _interopRequireDefault(_jquery);

var _backbone = require('backbone');

var _Socket = require('../utils/Socket');

var _Socket2 = _interopRequireDefault(_Socket);

var _2 = require('./');

var _app = require('../app');

var _app2 = _interopRequireDefault(_app);

var _ServerConfig = require('../models/ServerConfig');

var _ServerConfig2 = _interopRequireDefault(_ServerConfig);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/*
  The module is used to establish a connection with a server as well as monitor
  the state of any current connection.

  To establish a connection call connect().

  To get a snapshot of a current connection, call getCurrentConnection().

  To be informed of the state of a connection as it changes - If you initiated the
  connection (via a call to connect()),you can use the promise handlers (progress,
  done, fail) of the promise returned via the connect() call. Alternatively,
  particularly if you weren't the one that made the connect() call, you can bind
  to the { events } object exported in this module, which will fire relevant events
  ('connecting', 'connected', etc...).
*/

var events = _extends({}, _backbone.Events);

exports.events = events;


var getLocalServer = _underscore2.default.once(function () {
  return _electron.remote.getGlobal('localServer');
});

var serverCurStartArgMap = {
  BCH: '--bitcoincash',
  ZEC: '--zcash'
};

var defaultLocalServerStartArgs = function defaultLocalServerStartArgs() {
  var ls = getLocalServer();
  return ls && ls.lastStartCommandLineArgs || [];
};

/**
 * Will convert an array of command line arguments for the local server into the
 * crypto currency they start the server in.
 */
var serverStartArgsToCoin = function serverStartArgsToCoin() {
  var args = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : defaultLocalServerStartArgs();

  var walletCur = 'BTC';

  Object.keys(serverCurStartArgMap).forEach(function (cur) {
    if (args.indexOf(serverCurStartArgMap[cur]) !== -1) walletCur = cur;
  });

  return walletCur;
};

var currentConnection = null;
var debugLog = '';

function log(msg) {
  if (typeof msg !== 'string') {
    throw new Error('Please provide a message.');
  }

  if (!msg) return;

  var logMsg = '[SERVER-CONNECT] ' + msg + _os.EOL;
  debugLog += logMsg;
  // "if (ipcRenderer)" needed to prevent the `npm test` from bombing.
  if (_electron.ipcRenderer) _electron.ipcRenderer.send('server-connect-log', logMsg);
}

function getDebugLog() {
  return debugLog;
}

/**
 * Returns an object with data about the current connection, incuding
 * the status (e.g. 'connecting', 'connected'...) and any relevant data
 * (e.g. the Socket instance, the Server Configuration model...).
 */
function getCurrentConnection() {
  return currentConnection;
}

/**
 * Returns the server config model of the server that you are currently
 * connected to and null if you are not connected to a server.
 */
function getServer() {
  var curCon = getCurrentConnection();

  if (!curCon || !curCon.server || curCon.status !== 'connected') {
    return null;
  }

  return curCon.server;
}

/**
 * Call this method to obtain the socket instance in order to bind socket events.
 * If we are not currently connected to a server, this method will return false.
 * For the most part, you could no-op in that case since as it is now the
 * Connection Modal will overlay the app when there is no connection and any
 * subsequent reconnection will result in the app re-starting.
 */
function getSocket() {
  var serverConnection = getCurrentConnection();
  var returnVal = false;

  if (serverConnection && serverConnection.status !== 'disconnected') {
    returnVal = serverConnection.socket;
  }

  return returnVal;
}

function socketConnect(socket) {
  // todo: validate args

  var deferred = _jquery2.default.Deferred();

  var onOpen = function onOpen() {
    return deferred.resolve();
  };
  var onClose = function onClose(e) {
    return deferred.reject('failed', e);
  };

  var cancel = function cancel() {
    socket.close();
    socket.off(null, onOpen);
    socket.off(null, onClose);
    deferred.reject('canceled');
  };

  socket.on('open', onOpen);
  socket.on('close', onClose);
  socket.connect();

  var promise = deferred.promise();
  promise.cancel = cancel;

  return promise;
}

function authenticate(server) {
  // todo: validate args

  var deferred = _jquery2.default.Deferred();

  var fetchConfig = _jquery2.default.get(server.httpUrl + '/ob/config').done(function () {
    return deferred.resolve();
  }).fail(function (e) {
    return deferred.reject('failed', e);
  });

  var cancel = function cancel() {
    deferred.reject('canceled');
    fetchConfig.abort();
  };

  var promise = deferred.promise();
  promise.cancel = cancel;

  return promise;
}

/**
 * If we're currently connected to a server, this method will disconnect the connection.
 */
function disconnect() {
  var curCon = getCurrentConnection();
  if (curCon && curCon.socket) curCon.socket.close();
}

var _localServerStartHandlers = [];
var _proxySetHandlers = [];
var boundServerConfigRemove = false;

/**
 * Called to establish a connection with a server. This involves ensuring the
 * local server is running, if attempting to connect to a built-in
 * (i.e. local bundled) connection. It also involves opening a websocket connection
 * with the server. Additionally, if the configuration requires authentication,
 * it will call a server endpoint to make sure authentication was successful.
 *
 * @param {object} server - A ServerConfig instance representing the server you want to
 *   connect to.
 * @param {object} [options={}] - Options determining certain behavior of the connection
 *   process.
 * @param {number} [options.attempts=2] - The number of connection attempts to make. During
 *   the initial connection process (e.g. at startup), it is useful to have multiple attempts
 *   since the server may still be in the process of starting up.
 * @param {number} [options.minAttemptSpacing=3000] - The minimum number of milliseconds it
 *   should take from the start of one connection attempt until the start of the second on.
 *   This is very important because if a server is is down, in most cases the failure will
 *   be instant, if you don't enforce spacing all the attempts will happen in a fraction
 *   of a second, never giving a server a chance to startup. (Please note this governs the
 *   minimum spacing. The max is governed by maxAttemptTime)
 * @param {number} [options.maxAttemptTime=5000] - The maximum amount of milliseconds before
 *   giving up on a connection attempt (and moving on to the next one if there is one) and
 *   considering it a failure. This is particularly important if you are attempting to connect
 *   to a remote server with a non numeric ServerIp value. Since it's trying to get the name
 *   via DNS, it will take over a minute before it fails. The recommendation is to not make this
 *   value too large during start-up, or you risk making a very long start-up sequence,
 *   particularly if you have multiple attempts. In the Connection Management modal, the value
 *   could be much higher and the user could always cancel the attempt if they feel it is
 *   taking too long.
 * @return {object} A promise object will be returned that will allow you
 *   to monitor the progress and result of the connection via progress, done
 *   and fail handlers. The object will also include a cancel function, whereby
 *   you can cancel the connection attempt.
 */
function connect(server) {
  var options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

  if (!server instanceof _ServerConfig2.default) {
    throw new Error('Please provide a server as a ServerConfig instance.');
  }

  if (!boundServerConfigRemove) {
    boundServerConfigRemove = true;

    // If the config for the server we are connected to is removed, we'll
    // disconnect from the server.
    _app2.default.serverConfigs.on('remove', function (md) {
      var curCon = getCurrentConnection();
      if (curCon && curCon.server && curCon.server.id === md.id) {
        disconnect();
      }
    });
  }

  _app2.default.serverConfigs.activeServer = server;
  var curCon = getCurrentConnection();

  var innerLog = function innerLog() {
    var msg = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : '';

    log('[' + server.id.slice(0, 8) + '] ' + msg);
  };

  innerLog('Will attempt to connect to server "' + server.get('name') + '"' + (' at ' + server.get('serverIp') + '.'));

  var opts = _extends({
    attempts: 10,
    minAttemptSpacing: 3500,
    maxAttemptTime: 5000
  }, options);

  var deferred = _jquery2.default.Deferred();
  var localServer = getLocalServer();
  var serverCurrency = server.get('walletCurrency');
  var attempt = 1;
  var socket = null;
  var connectAttempt = null;
  var cancel = null;

  var getPromiseData = function getPromiseData() {
    var data = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
    var getPromiseOpts = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

    var getPromiseDataOpts = _extends({
      includeAttemptData: true
    }, getPromiseOpts);

    var aggregateData = _extends({
      localServer: localServer,
      server: server,
      socket: socket
    }, data);

    if (getPromiseDataOpts.includeAttemptData) {
      aggregateData = _extends({}, aggregateData, {
        connectAttempt: attempt,
        totalConnectAttempts: opts.attempts
      });
    }

    if (localServer) aggregateData.localServer = localServer;

    return aggregateData;
  };

  var notify = function notify(e) {
    currentConnection = _extends({}, getPromiseData(e), {
      cancel: cancel
    });

    events.trigger(e.status, getPromiseData(e));
    deferred.notify(getPromiseData(e));
  };

  var reject = function reject(e) {
    currentConnection = _extends({}, getPromiseData(e), {
      status: 'connect-attempt-failed'
    });

    events.trigger('connect-attempt-failed', getPromiseData(e));
    deferred.reject(getPromiseData(e));
  };

  var resolve = function resolve(e) {
    var data = getPromiseData(e);

    currentConnection = _extends({}, data, {
      status: 'connected'
    });

    data.socket.on('close', function (connectionLostE) {
      var connectionLostEventData = getPromiseData({ socketCloseEvent: connectionLostE }, { includeAttemptData: false });
      events.trigger('disconnected', connectionLostEventData);
      if (!currentConnection || currentConnection.socket === data.socket) {
        currentConnection = _extends({}, data, {
          status: 'disconnected'
        });
      }
    });

    events.trigger('connected', getPromiseData(e));
    deferred.resolve(getPromiseData(e));
  };

  cancel = function cancel() {
    if (connectAttempt) connectAttempt.cancel();
  };

  if (server.get('builtIn') && !localServer) {
    // This should never happen to normal users. The only way it would is if you are a dev
    // and mucking with localStorage and / or fudging the source for the app to masquerade
    // as a bundled app.
    throw new Error('A configuration for a built-in server should only be used on ' + ' the bundled app.');
  }

  var curLocalServerCoin = serverStartArgsToCoin();
  var curLocalServerZecBinaryPath = null;

  if (localServer) {
    if (curLocalServerCoin === 'ZEC') {
      var zecIndex = localServer.lastStartCommandLineArgs.indexOf(serverCurStartArgMap.ZEC);

      if (zecIndex !== -1 && typeof localServer.lastStartCommandLineArgs[zecIndex + 1] === 'string') {
        curLocalServerZecBinaryPath = localServer.lastStartCommandLineArgs[zecIndex + 1];
      }
    }
  }

  // If we're not connecting to the local bundled server or it's running with incompatible
  // command line arguments, let's ensure it's stopped.
  if (localServer && localServer.isRunning && (!server.get('builtIn') || serverCurrency !== serverStartArgsToCoin() || serverCurrency === 'ZEC' && server.get('zcashBinaryPath') !== curLocalServerZecBinaryPath)) {
    deferred.notify({ status: 'stopping-local-server' });
    localServer.stop();

    if (server.get('builtIn') && options.attempts === undefined && options.minAttemptSpacing === undefined && options.maxAttemptTime === undefined) {
      // If we need to wait for a bundled server to shut down before starting a new instance and
      // the user did not pass in override options regarding connection timeouts, we'll bump up the
      // defaults to give the shutting down server more time (it may take up to a few minutes).
      opts = _extends({}, opts, {
        attempts: 60, // works out to 5 minutes total
        minAttemptSpacing: 5000,
        maxAttemptTime: 5000
      });
    }
  }

  if (curCon) {
    if (curCon.cancel) curCon.cancel();
    disconnect();
  }

  socket = new _Socket2.default(server.socketUrl);

  var innerConnect = function innerConnect() {
    var innerConnectDeferred = _jquery2.default.Deferred();

    var socketConnectAttempt = null;
    var connectCancel = function connectCancel(reason) {
      innerConnectDeferred.reject(reason || 'canceled');
      if (socketConnectAttempt) socketConnectAttempt.cancel();
    };

    // The next three are just wrappers around actions on the innerConnectDeferred
    // so they are done in a timeout because we want the promise from this function to
    // be returned first.
    var innerConnectResolve = function innerConnectResolve() {
      for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
        args[_key] = arguments[_key];
      }

      setTimeout(function () {
        return innerConnectDeferred.resolve.apply(innerConnectDeferred, args);
      });
    };

    var innerConnectReject = function innerConnectReject() {
      for (var _len2 = arguments.length, args = Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
        args[_key2] = arguments[_key2];
      }

      setTimeout(function () {
        return innerConnectDeferred.reject.apply(innerConnectDeferred, args);
      });
    };

    var innerConnectNotify = function innerConnectNotify() {
      for (var _len3 = arguments.length, args = Array(_len3), _key3 = 0; _key3 < _len3; _key3++) {
        args[_key3] = arguments[_key3];
      }

      setTimeout(function () {
        return innerConnectDeferred.notify.apply(innerConnectDeferred, args);
      });
    };

    innerConnectNotify('connecting');

    // If we're connecting to the built in server and the existing local server instance is
    // stopping, we'll do nothing until the existing instance stops completely (may be a minute
    // or two)
    var willWaitForLocalServerStop = false;

    if (server.get('builtIn') && localServer.isStopping) {
      willWaitForLocalServerStop = true;
      var stoppingServer = _app2.default.serverConfigs.findWhere({ walletCurrency: serverStartArgsToCoin() });
      innerConnectNotify('waiting-for-local-server-stop', { stoppingServer: stoppingServer });
      innerLog('Waiting for the local server started with ' + ('"' + localServer.lastStartCommandLineArgs + '" to stop.'));
    } else {
      if (willWaitForLocalServerStop) innerConnectNotify('local-server-stopped');

      // This is called when either the client Tor proxy has been set or once we've determined
      // it's not needed.
      var onClientTorProxyChecked = function onClientTorProxyChecked() {
        // This flag means that we want the server to be running in a certain tor mode
        // (tor on or tor off), buts it's already running in the opposite mode.
        var serverRunningIncompatibleWithTor = server.get('builtIn') && localServer.isRunning && server.get('useTor') !== (localServer.lastStartCommandLineArgs.indexOf('--tor') !== -1);

        if (serverRunningIncompatibleWithTor) {
          // The idea is you will probably be starting the server with multiple attempts. So
          // if we stop the server now, on a subsequent attempt we'll re-start it in the
          // desired mode.
          localServer.stop();
          innerConnectReject('incompatible-tor-mode');
        }

        if (server.get('builtIn') && !localServer.isRunning) {
          var onTorChecked = function onTorChecked() {
            var onLocalServerStart = function onLocalServerStart() {
              socketConnectAttempt = socketConnect(socket).done(function () {
                innerConnectResolve('connected');
              }).fail(function (reason, e) {
                if (reason === 'canceled') {
                  innerConnectReject('canceled');
                } else {
                  innerConnectReject('socket-connect-failed', { socketCloseEvent: e });
                }
              });
            };

            var commandLineArgs = ['-v'];

            if (serverCurrency !== 'BTC') {
              var serverCoinArg = serverCurStartArgMap[serverCurrency];

              if (serverCoinArg) {
                commandLineArgs.push(serverCoinArg);

                if (serverCurrency === 'ZEC') {
                  commandLineArgs.push(server.get('zcashBinaryPath'));
                }
              }
            }

            if (server.get('useTor')) commandLineArgs.push('--tor');
            var torPw = server.get('torPassword');
            if (torPw) {
              commandLineArgs = commandLineArgs.concat(['--torpassword', torPw]);
            }
            innerConnectNotify('starting-local-server');
            localServer.start(commandLineArgs);

            // Remove any previous start handlers that this module may have bound. Not
            // removing them all, because other modules bind to 'start' and we don't
            // want to remove their handlers.
            _localServerStartHandlers.forEach(function (handler) {
              return localServer.off('start', handler);
            });
            _localServerStartHandlers = [onLocalServerStart];

            localServer.on('start', onLocalServerStart);
          };

          if (server.get('confirmedTor') && !server.get('useTor')) {
            onTorChecked();
          } else {
            var getServerStatusPid = localServer.getServerStatus();

            localServer.on('getServerStatusSuccess', function (data) {
              if (data.pid === getServerStatusPid) {
                if (data.torAvailable && !server.get('useTor')) {
                  innerConnectReject('tor-not-configured');
                } else if (!data.torAvailable && server.get('useTor')) {
                  innerConnectReject('tor-not-available');
                } else {
                  onTorChecked();
                }
              }
            });

            localServer.on('getServerStatusFail', function (data) {
              if (data.pid === getServerStatusPid) {
                innerConnectReject('unable-to-get-server-status');
              }
            });
          }
        } else {
          socketConnectAttempt = socketConnect(socket).done(function () {
            if (server.needsAuthentication()) {
              innerConnectNotify('authenticating');
              authenticate(server).done(function () {
                return innerConnectResolve('connected');
              }).fail(function (reason, e) {
                innerConnectReject('authentication-failed', { failedAuthEvent: e });
              });
            } else {
              innerConnectResolve('connected');
            }
          }).fail(function (reason, e) {
            if (reason === 'canceled') {
              innerConnectReject('canceled');
            } else {
              innerConnectReject('socket-connect-failed', { socketCloseEvent: e });
            }
          });
        }
      };

      if (server.get('useTor')) {
        innerConnectNotify('setting-tor-proxy');
        innerLog('Activating a proxy at socks5://' + server.get('torProxy'));
        var setProxyId = (0, _2.guid)();

        var onProxySet = function onProxySet(e, id) {
          if (id === setProxyId) {
            innerConnectNotify('tor-proxy-set');
            onClientTorProxyChecked();
          }
        };

        _electron.ipcRenderer.send('set-proxy', setProxyId, 'socks5://' + server.get('torProxy'));
        _proxySetHandlers.forEach(function (handler) {
          return _electron.ipcRenderer.removeListener('proxy-set', handler);
        });
        _electron.ipcRenderer.on('proxy-set', onProxySet);
        _proxySetHandlers = [onProxySet];
      } else {
        innerConnectNotify('clearing-tor-proxy');
        innerLog('Clearing any proxy that may be set.');
        var _setProxyId = (0, _2.guid)();
        _electron.ipcRenderer.send('set-proxy', _setProxyId, '');

        var _onProxySet = function _onProxySet(e, id) {
          if (id === _setProxyId) {
            innerConnectNotify('tor-proxy-cleared');
            onClientTorProxyChecked();
          }
        };

        _proxySetHandlers.forEach(function (handler) {
          return _electron.ipcRenderer.removeListener('proxy-set', handler);
        });
        _electron.ipcRenderer.on('proxy-set', _onProxySet);
        _proxySetHandlers = [_onProxySet];
      }
    }

    var promise = innerConnectDeferred.promise();
    promise.cancel = connectCancel;

    return promise;
  };

  var connectAttemptStartTime = null;
  var attemptConnection = function attemptConnection() {
    var maxTimeTimeout = null;
    var nextAttemptTimeout = null;
    connectAttemptStartTime = Date.now();

    var innerConnectAttempt = innerConnect().progress(function (status) {
      var data = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
      return notify(_extends({ status: status }, data));
    }).done(function (status) {
      var data = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
      return resolve(_extends({ status: status }, data));
    }).fail(function (status) {
      var data = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

      if (attempt === opts.attempts || status === 'authentication-failed' || status === 'outer-connect-attempt-canceled' || status === 'tor-not-configured' || status === 'tor-not-available') {
        var reason = void 0;

        if (status === 'socket-connect-failed') {
          reason = 'unable-to-reach-server';
        } else if (status === 'outer-connect-attempt-canceled') {
          reason = 'canceled';
        } else {
          reason = status;
        }

        reject(_extends({
          reason: reason
        }, data));
      } else {
        var delay = opts.minAttemptSpacing - (Date.now() - connectAttemptStartTime);

        nextAttemptTimeout = setTimeout(function () {
          attempt += 1;
          connectAttempt = attemptConnection();
        }, delay < 0 ? 0 : delay);
      }
    }).always(function () {
      return clearTimeout(maxTimeTimeout);
    });

    var attemptConnectionCancel = function attemptConnectionCancel() {
      reject({ reason: 'canceled' });
      clearTimeout(maxTimeTimeout);
      clearTimeout(nextAttemptTimeout);
      innerConnectAttempt.cancel('outer-connect-attempt-canceled');
    };

    maxTimeTimeout = setTimeout(function () {
      if (innerConnectAttempt.state() === 'pending') innerConnectAttempt.cancel('timed-out');
    }, opts.maxAttemptTime);

    return { cancel: attemptConnectionCancel };
  };

  connectAttempt = attemptConnection();

  // wire in some logging
  deferred.progress(function (e) {
    innerLog('Status is "' + e.status + '" for connect attempt' + (' ' + e.connectAttempt + ' of ' + e.totalConnectAttempts + '.'));
  }).done(function (e) {
    innerLog('Connected to "' + e.server.get('name') + '"');

    e.socket.on('close', function () {
      innerLog('Disconnected from "' + e.server.get('name') + '"');
    });
  }).fail(function (e) {
    innerLog('Failed to connect to "' + e.server.get('name') + '"');
    innerLog('====> Reason: ' + e.status);

    if (e.socketCloseEvent) {
      innerLog('====> Code: ' + e.socketCloseEvent.code);
      if (e.socketCloseEvent.reason) log('Reason: ' + e.socketCloseEvent.reason);
    } else if (e.failedAuthEvent) {
      innerLog('====> Status: ' + e.failedAuthEvent.status);
      innerLog('====> Status text: ' + e.failedAuthEvent.statusText);

      if (e.failedAuthEvent.responseText) {
        innerLog('====> Response text: ' + e.failedAuthEvent.responseText);
      }
    }
  });

  var promise = deferred.promise();
  promise.cancel = cancel;

  return promise;
}

// "if (ipcRenderer)" needed to prevent the `npm test` from bombing.
if (_electron.ipcRenderer) _electron.ipcRenderer.send('server-connect-ready');
log('Browser has been started or refreshed.');